import logging
from io import StringIO

logging.basicConfig(format="%(levelname)-8s %(message)s", level=logging.INFO)


def init_logger():
    stream = StringIO()
    console = logging.StreamHandler(stream)
    console.setFormatter(logging.Formatter("%(levelname)-12s %(message)s"))
    weld_logger = logging.getLogger("weld")
    weld_logger.addHandler(console)
    weld_logger.setLevel(logging.INFO)

    return stream



import time
from importlib import resources
from pathlib import Path
from zipfile import ZipFile

import streamlit as st
import yaml
from beet import WrappedException
from streamlit.delta_generator import DeltaGenerator
from streamlit_extras.stateful_button import button as toggle_button

from smithed import weld

from .log_helpers import init_logger
from .models import WebApp

icon = "https://github.com/Smithed-MC/smithed-python/blob/main/smithed/weld/resources/icon.png?raw=true"

webapp = WebApp.parse_obj(
    yaml.safe_load(
        (resources.files("smithed") / "weld/resources/webapp.yaml").read_text()
    )
)

# def validate_zips(packs: list[str]):
#     for pack in packs:
#         try:
#             with pack.open("pack.mcmeta") as meta:
#                 print(meta.read(), "\n")
#         except FileNotFoundError:
#             return f"`pack.mcmeta` not found in pack {pack.filename}"

#     return False


def upload_flow(ui: DeltaGenerator):
    progress = ui.container()
    raw_packs = ui.file_uploader("Upload packs", accept_multiple_files=True, type="zip")
    packs = [ZipFile(pack) for pack in raw_packs]

    col1, col2, col3 = ui.columns(3)
    with col2:
        if fabric_mod := toggle_button("Turn into Fabric Mod", key="toggle"):
            ui.warning(webapp.fabric)

    path = None
    t0 = time.perf_counter()
    if col1.button("Build Packs", disabled=not packs, key="build") and packs:
        # for pack in packs:
        #     with (temp / pack.name).with_suffix(".zip").open("wb") as temp_file:
        #         temp_file.write(pack.read())
        # pack_paths = list(str(path) for path in temp.glob("*.zip"))
        # if error := validate_zips(pack_paths):
        #     st.error(error)
        #     return
        with st.status(f"Welding {len(packs)} packs!", expanded=False) as status:
            stream = init_logger()
            stream.seek(0)
            stream.truncate()
            try:
                with weld.run_weld(packs, as_fabric_mod=fabric_mod) as ctx:
                    if fabric_mod:
                        with ZipFile("output.jar", "w") as jar:
                            ctx.data.dump(jar)
                            ctx.assets.dump(jar)
                            path = Path(str(jar.filename))
                    else:
                        if ctx.data:
                            path = ctx.data.save(
                                path="output", zipped=True, overwrite=True
                            )
                        if ctx.assets:
                            path = ctx.assets.save(
                                path="output", zipped=True, overwrite=True
                            )

                t1 = time.perf_counter()
                stream.seek(0)
                st.code(stream.getvalue())
                status.update(
                    label=f"Merged in :green[{t1 - t0: 0.3f}s]. Click to see log."
                )

            except WrappedException as exc:
                status.update(label=":red[Error occured. Click to reveal error.]")
                st.exception(exc.__cause__)

        if path is not None:
            with path.open(mode="rb") as file:
                col3.download_button(
                    "Download Welded Packs",
                    file,
                    file_name="welded-pack.zip" if not fabric_mod else "welded-mod.jar",
                )
    else:
        col3.button("Download Welded Packs", disabled=True, key="download")

    if not packs:
        progress.info("Waiting for uploaded packs")

    elif path is None:
        progress.success("Ready to Build!")


def main():
    st.set_page_config(page_title="Weld", page_icon=icon, layout="wide")

    st.sidebar.title("Special Merging Rules")
    st.sidebar.write(webapp.conflicts)
    st.sidebar.write("----")
    st.sidebar.write("See [docs](https://wiki.smithed.dev/weld) for more info!")

    st.write(webapp.title.format(version=weld.__version__), unsafe_allow_html=True)

    st.write(webapp.intro, unsafe_allow_html=True)

    st.warning(webapp.warn)
    upload_flow(st.container())

    st.write("---")

    col1, col2 = st.columns(2)
    col1.markdown(webapp.footer.left, unsafe_allow_html=True)
    col2.markdown(webapp.footer.right, unsafe_allow_html=True)



import logging
import shutil
from collections.abc import Sequence
from glob import glob
from pathlib import Path
from time import perf_counter
from typing import Annotated, Optional

import typer
from beet import ProjectCache
from rich.logging import RichHandler
from rich.panel import Panel

from smithed.theming import console, print
from smithed.utils import error_handler, path_str

from .main import run_weld

cli = typer.Typer(
    help="🔧 Weld multiple data and resource packs into a single zip!",
    add_completion=False,
    pretty_exceptions_enable=True,
    pretty_exceptions_show_locals=False,
    rich_markup_mode="markdown",
)

logger = logging.getLogger("weld")
OUTPUT_DIR, OUTPUT = "output", "welded-pack.zip"


def version_callback(value: bool):
    if not value:
        return
    from smithed import weld

    print("weld", style="secondary", end=" ")
    print(f"v{weld.__version__}", highlight=False, style="accent_light")
    raise typer.Exit()


@cli.command(
    epilog="Made with :heart: by [Smithed](https://beta.smithed.dev)",
    no_args_is_help=True,
)
def weld(
    packs: Annotated[list[str], typer.Argument(help="A series of packs to weld")],
    dir: Annotated[Path, typer.Option(help="Output directory")] = Path(OUTPUT_DIR),
    name: Annotated[Path, typer.Option(help="Output file name")] = Path(OUTPUT),
    log: Annotated[str, typer.Option(help="Log level")] = "INFO",
    clear_cache: Annotated[bool, typer.Option(help="Clear cache")] = False,
    dev: Annotated[bool, typer.Option(help="Enable dev mode")] = False,
    version: Annotated[
        Optional[bool],
        typer.Option("--version", callback=version_callback, is_eager=True),
    ] = None,
):
    """Weld data and resource packs together with all data and assets in one zip.

    See the [docs](https://wiki.smithed.dev/weld) more info!
    """

    setup_logger(dev, log.upper())

    packs = list(expand_globs(packs))
    if len(packs) < 1:
        raise typer.BadParameter("Need at least one pack to weld")

    weld_config_dir = Path(typer.get_app_dir("smithed")) / "weld"
    cache = weld_config_dir / "beet_cache"

    if clear_cache:
        shutil.rmtree(cache)

    print(
        Panel(
            "[bold][accent] Weld [/bold]by Smithed ",
            border_style="accent_light",
        ),
    )

    success = False

    logger.debug("Using cache: '%s'", path_str(cache))
    logger.info("Packs: %s", ", ".join(f"'{pack}'" for pack in packs))

    t0 = perf_counter()
    with (
        error_handler(should_exit=True),
        run_weld(packs, cache=ProjectCache(cache, cache)) as ctx,
    ):
        logger.info(
            "[success]Saving[/success], might take a while.",
            extra={"markup": True},
        )
        ctx.data.save(path=dir / name, zipped=name.suffix == ".zip", overwrite=True)
        success = True
    t1 = perf_counter()

    if success:
        print()
        print(
            Panel(
                f"[success][bold]✔️[/bold] Welded [bold]{len(packs)}[/bold] packs"
                f" in [bold]{t1-t0:.3f}s[/bold]"
                f"\n[success]Output: [bold]{path_str((dir / name).resolve())}[/bold]",
                border_style="accent_light",
            )
        )
    else:
        print()
        print(Panel("[error]:boom: Failed to weld packs", border_style="error"))
        raise typer.Abort()


def setup_logger(dev: bool, log_level: str):
    if dev:
        log_level = "DEBUG"

    logging.basicConfig(
        level=log_level,
        format="%(message)s",
        datefmt="[%X]",
        handlers=[RichHandler(console=console, rich_tracebacks=False, show_path=dev)],
    )


def expand_globs(packs: Sequence[str]):
    for pack in packs:
        yield from glob(pack)


if __name__ == "__main__":
    typer.run(weld)



from .cli import cli
from .helper_plugins import inject_pack_id_into_smithed, print_pack_name
from .main import run_weld

__all__ = [
    "cli",
    "run_weld",
    "print_pack_name",
    "inject_pack_id_into_smithed",
]



import logging
import os
from pathlib import Path

import pytest
from lectern import Document
from pytest_insta import SnapshotFixture

from smithed.weld import run_weld

EXAMPLES = [f for f in os.listdir("examples") if not f.startswith(".")]

TEST_CONFIG = {
    "require": ["beet.contrib.auto_yaml"],
}


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
@pytest.mark.parametrize("directory", EXAMPLES)
def test_build(
    snapshot: SnapshotFixture, directory: str, caplog: pytest.LogCaptureFixture
):
    packs = sorted(
        [
            f"examples/{directory}/{pack.name}"
            for pack in (Path("examples") / directory).glob("*")
            if pack.is_dir()
        ]
    )
    with caplog.at_level(logging.WARNING), run_weld(packs, config=TEST_CONFIG) as ctx:
        document = ctx.inject(Document)
        document.markdown_serializer.flat = True
        for record in caplog.records:
            if record.levelname == "ERROR":
                raise AssertionError("Logger revealed error")
        assert snapshot("pack.md") == document



import logging
from typing import Literal

from beet import ListOption
from pydantic import validator

from .base import BaseModel

logger = logging.getLogger("weld")


class Priority(BaseModel):
    stage: Literal["early", "standard", "late"] = "standard"
    default: int = 0

    before: ListOption[str] = ListOption()
    after: ListOption[str] = ListOption()

    @validator("before", "after")
    def convert_fields(cls, value: ListOption[str]):
        return ListOption(__root__=list(dict.fromkeys(value.entries())))



from typing import TypeVar

JsonType = int | str | float | bool | None | list["JsonType"] | dict[str, "JsonType"]
JsonDict = dict[str, "JsonType"]
JsonList = list["JsonType"]

JsonTypeT = TypeVar("JsonTypeT", bound=JsonType)
